# coding=utf-8
# Copyright 2018-2022 EVA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

import pytest
from test.util import (
    DummyObjectDetector,
    create_sample_video,
    file_remove,
    load_udfs_for_testing,
    shutdown_ray,
)

import pandas as pd

from eva.catalog.catalog_manager import CatalogManager
from eva.configuration.constants import EVA_ROOT_DIR
from eva.executor.executor_utils import ExecutorError
from eva.models.storage.batch import Batch
from eva.server.command_handler import execute_query_fetch_all

NUM_FRAMES = 10


class CreateTableTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # reset the catalog manager before running each test
        CatalogManager().reset()
        video_file_path = create_sample_video()
        load_query = f"LOAD VIDEO '{video_file_path}' INTO MyVideo;"
        execute_query_fetch_all(load_query)
        ua_detrac = f"{EVA_ROOT_DIR}/data/ua_detrac/ua_detrac.mp4"
        execute_query_fetch_all(f"LOAD VIDEO '{ua_detrac}' INTO UATRAC;")
        load_udfs_for_testing()

    @classmethod
    def tearDownClass(cls):
        shutdown_ray()
        file_remove("dummy.avi")
        file_remove("ua_detrac.mp4")
        execute_query_fetch_all("DROP TABLE IF EXISTS MyVideo;")
        execute_query_fetch_all("DROP TABLE IF EXISTS UATRAC;")

    def setUp(self):
        execute_query_fetch_all("DROP TABLE IF EXISTS dummy_table;")
        execute_query_fetch_all("DROP TABLE IF EXISTS uadtrac_fastRCNN;")

    def tearDown(self):
        execute_query_fetch_all("DROP TABLE IF EXISTS dummy_table;")
        execute_query_fetch_all("DROP TABLE IF EXISTS uadtrac_fastRCNN;")

    def test_currently_cannot_create_boolean_table(self):
        query = """ CREATE TABLE BooleanTable( A BOOLEAN);"""

        with self.assertRaises(ExecutorError):
            execute_query_fetch_all(query)

    def test_should_create_table_from_select(self):
        create_query = """CREATE TABLE dummy_table
            AS SELECT id, DummyObjectDetector(data).label FROM MyVideo;
        """
        execute_query_fetch_all(create_query)

        select_query = "SELECT id, label FROM dummy_table;"
        actual_batch = execute_query_fetch_all(select_query)
        actual_batch.sort()

        labels = DummyObjectDetector().labels
        expected = [
            {"dummy_table.id": i, "dummy_table.label": [labels[1 + i % 2]]}
            for i in range(NUM_FRAMES)
        ]
        expected_batch = Batch(frames=pd.DataFrame(expected))
        self.assertEqual(actual_batch, expected_batch)

    @pytest.mark.torchtest
    def test_should_create_table_from_select_lateral_join(self):
        select_query = (
            "SELECT id, label, bbox FROM UATRAC JOIN LATERAL "
            "Yolo(data) AS T(label, bbox, score) WHERE id < 5;"
        )
        query = (
            "CREATE TABLE IF NOT EXISTS "
            f"uadtrac_fastRCNN AS {select_query};"
        )
        execute_query_fetch_all(query)

        select_view_query = "SELECT id, label, bbox FROM uadtrac_fastRCNN"
        actual_batch = execute_query_fetch_all(select_view_query)
        actual_batch.sort()

        self.assertEqual(len(actual_batch), 5)
        # non-trivial test case
        res = actual_batch.frames
        for idx in res.index:
            self.assertTrue("car" in res["uadtrac_fastrcnn.label"][idx])
